-- ox.ex
-- converts an LL(1) grammar file into an executable
-- euphoria program.

constant
    AppName         = "Ox: A Parser Generator",
    AppVersion      = "v3.2",
    AppCopyright    = "(c) David Cuny, 2001"
    

with trace

include string.e
include get.e
include sort.e
include wildcard.e

constant
    False   = 0,
    True    = not False

-- variables controlling syntax
constant
    SyntaxVars = {
        "AbortOnError",
        "IgnoreCase",
        "InitKeyWordChars",
        "KeyWordChars",
        "CommentStart",
        "CommentEnd",
        "SingleComment",
        "SingleComment2",
        "StringStart",
        "StringEnd",
        "StringEsc",
        "CharStart",
        "CharEnd",
        "CharEsc",
        "HexStart" }
    

integer 
    inFile,         -- ox file being read
    outFile,        -- file being created
    lineNum,        -- current line number in file
    inCode,         -- true if parsing code           
    inRule,         -- current rule number
    ignoreCase       -- true if keywords are set to upper case
    
sequence
    fileName,       -- name of file being parsed
    lineText,       -- text from file
    fileText,       -- file
    grammar,        -- holds pattern rule
    code,           -- holds code pre- and post- code associated with rule
    syntaxList,     -- list of syntax values
    keywordList,    -- list of keywords
    tokenList,      -- list of all declared tokens
    ruleList        -- list of all rules seen (not declared)
    
-- these are used by string_to_words
constant
    Whitespace  = " \t"

-- these are used by replaceid    
constant 
    IdChar =    "1234567890" &
                "QWERTYUIOPASDFGHJKLZXCVBNM" &
                "qwertyuiopasdfghjklzxcvbnm" &
                "_"

-- the sections of an ox file
constant
    SyntaxSection   = 1,
    GrammarSection  = 2

-- what type of code is being parsed
constant
    Pre     = 1,
    Post    = 2

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- General Tools
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

procedure write( object o )
    -- for debugging
    if integer(o) then
        printf( 1, "%d'%s'", {o,o} )
    else
        puts( 1, "{" )
        for i = 1 to length( o ) do
            write( o[i] )
            if i < length( o ) then
                puts( 1, "," )
            end if
        end for
        puts( 1, "}" )
    end if
end procedure



-----------------------------------------------------------------------------
procedure fatal( sequence text, sequence args )
    -- print error message, and shut down
    puts( 1, "\n" )
    printf( 1, "Error in line %d of file %s\n", {lineNum, fileName} )
    puts( 1, fileText[lineNum] & '\n' )
    printf( 1, text & "\n", args )
    abort(0)
end procedure

-----------------------------------------------------------------------------
procedure tab( integer depth )
    puts( outFile, repeat( ' ', (depth)*4 ) )
end procedure


-----------------------------------------------------------------------------
function group( sequence s )
    -- group all sequence items with common first elements
    integer at
    sequence heads, group
    object head
    

    heads = {}
    group = {}
    
    -- build list of heads
    for i = 1 to length( s ) do
        head = s[i][1]
        at = find( head, heads )
        if at then
            group[at] &= {s[i]}
        else
            heads &= {head}
            group &= {{s[i]}}
        end if
    end for
    return group
end function

-----------------------------------------------------------------------------
function remove( sequence s, integer i )
    return s[1..i-1] & s[i+1..length(s)]
end function

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- The OX File Parser
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------





-----------------------------------------------------------------------------
function string_to_words( sequence string )
    
    -- convert string to words
    sequence words, word

    words = {}
    word = ""
    string &= " "
    for i = 1 to length( string ) do
    
        -- found a delimiter?
        if find( string[i], Whitespace ) then
            -- word accumulated?
            if length( word ) then
                words &= {word}
                word = ""
            end if
            
            -- treat delimiter as word?
            if string[i] != ' ' then
                words &= {{string[i]}}
            end if
            
        else               
            -- accumulate
            word &= string[i]
        end if
    end for

    return words

end function

integer indent
-----------------------------------------------------------------------------
procedure parseCode( sequence text )

    -- parse the code from the line
    integer at              

    -- pre code tag?
    if not inCode then   
        at = match( "<:", text )
        if at then
            inCode = Pre
            indent = at+3
        end if
    end if

    -- post code tag?
    if not inCode then   
        at = match( "{:", text )
        if at then
            inCode = Post
            indent = at+3
        end if
    end if
    
    if inCode then         
        -- look for end tag
        if inCode = Pre then
            at = match( ":>", text )
        else
            at = match( ":}", text )
        end if
        
        if at then        
            -- add to code
            code[inRule][inCode] &= "\n   " & text[indent..at-1]
            
            -- clear flag
            inCode = False
        else              
            -- add to code
            code[inRule][inCode] &= "\n   " & text[indent..length(text)]
        end if
    end if
    
    
end procedure

-----------------------------------------------------------------------------
procedure parseRule( sequence words )
    -- parse a rule
    sequence thisRule, token

    -- skip if blank
    if length( words ) = 0 then
        return
    end if

    -- clear rule
    thisRule = {}
    
    -- new grammar
    grammar = append( grammar, {} )

    -- set inRule
    inRule = length( grammar )

    -- continuation of existing rule?
    if equal( words[1], "|" ) then                              
        -- get head of prior rule
        thisRule = append( thisRule, grammar[inRule-1][1] )
        
    elsif equal( words[2], "=" ) then
        -- prepend RUL
        thisRule = append( thisRule, "RUL" & words[1] )

        -- remove second element
        words = remove( words, 2 )
        
    else
        fatal( "Expected start of grammar rule, not %s.", {words[1]} )
    end if

    -- convert remaining rules
    for i = 2 to length( words ) do
    
        -- get a word
        token = words[i]

        -- quoted?
        if token[1] = '\'' then     
        
            -- remove end quote
            if token[length(token)]  != '\'' then
                fatal( "No closing quote on %s", {token})
            else
                token = token[2..length(token)-1]
            end if

            -- ignore case?
            if ignoreCase then
                token = upper( token )
            end if
            
            -- not in the token list?
            if not find( token, tokenList ) then
                -- add to the token list
                tokenList &= { token }
                
                -- add to the keyword list
                keywordList &= { token }
                
            end if

            -- prepend with literal prefix
            token = "TOK" & token     

        -- pre-processing flag?
        elsif equal( token, "$pre" ) then
            -- pre-action routine
            token = "PRE"
                        
        -- special char?
        elsif not find( token[1], IdChar ) then
            fatal( "Expected %s to be in single quotes i.e.: '%s'",
                    {token, token} )

        -- token?
        elsif equal( upper(token), token ) then

            -- ignore case?
            if ignoreCase then
                token = upper( token )
            end if
            
            -- need to add to token list?
            if not find( token, tokenList ) then
                -- add to token list
                tokenList &= { token }
            end if
                            
            -- prepend with token prefix
            token = "TOK" & token

        -- must be a rule            
        else
            
            -- add to ruleList
            if not find( token, ruleList ) then
                ruleList &= { token }
            end if       
            
            -- prepend with rule prefix
            token = "RUL" & token
            
        end if

        -- append
        thisRule = append( thisRule, token )
        
    end for

    -- save    
    grammar[inRule] = thisRule
    
    -- set pre and post code sections blank
    code = append( code, { "", "" } )
        
end procedure     
                   
-----------------------------------------------------------------------------
function replace_id( sequence s, sequence source, sequence target )

    -- replace occurances of source with target
    -- this differs from replace() in that the source string must
    -- be surrounded by delimiters
    
    integer at, frontDelim, rearDelim
    sequence changed

    -- this accumulates the result    
    changed = ""
    while 1 do                 
    
        -- look for source string
        at = match( source, s )
        
        -- no match?
        if at = 0 then
            -- leave loop
            exit
        end if
        
        -- delimiter in front?
        if at = 1 then
            -- treat as delimiter
            frontDelim = True
        else         
            -- return true if it's not a valid id char
            frontDelim = not find( s[at-1], IdChar )
        end if
           
        -- delimiter in back?
        if at+length(source) > length(s) then
            -- treat as delimiter
            rearDelim = True
        else         
            -- return true if it's not a valid id char
            rearDelim = not find( s[at+length(source)], IdChar )
        end if

        -- not delimited?           
        if not frontDelim
        or not rearDelim then
            -- move to changed
            changed &= s[1..at+length(source)-1]
            s = s[at+length(source)..length(s)]
        else
            -- move everything prior to the match to changed string
            changed &= s[1..at-1]
    
            -- add the replacement text to the changed string
            changed &= target
            
            -- remove everything before the end from the string
            s = s[at+length(source)..length(s)]
            
        end if
        
    end while

    -- return the resulting string
    return changed & s

end function


-----------------------------------------------------------------------------
function stripToken( sequence s )
    -- return the token, stripped of the 3 character prefix
    if equal( s, "PRE" ) then
        return "$pre"
    else
        return right( s, -3 )
    end if
end function

-----------------------------------------------------------------------------
function tokenType( sequence token, sequence s )
    -- return true if type of token matches s
    -- compares type with three character prefix
    return equal( left( token, 3 ), s )
end function


-----------------------------------------------------------------------------
procedure writeRule( integer handle, integer index )

    -- write the rule 
    
    -- rule number
    printf( handle, "Rule #%d ==> ", {index} )    
    
    -- head of rule 
    puts( handle, stripToken(grammar[index][1]) & ": " )
    
    -- body of rule
    for i = 2 to length( grammar[index] ) - 1 do
        puts( handle, stripToken(grammar[index][i]) & " " )
    end for
    
end procedure

                   
-----------------------------------------------------------------------------
procedure writeRuleCode()
    -- generate functions to hanldle each rule
    sequence thisCode
    
    puts( outFile, "-------------------------------------\n" )
    puts( outFile, "-- Grammar Rule Code Section\n" )
    puts( outFile, "-------------------------------------\n" )     

    for i = 1 to length( code ) do

        for j = Pre to Post do
        
            -- get code
            thisCode = code[i][j]
    
            -- any code?
            if length( thisCode ) then
    
                -- replace $0 with s
                thisCode = replace_id( thisCode, "$0", "s" )
            
                -- replace $1 - $20 with s[1] - s[20]
                for k = 1 to 20 do
                    thisCode = replace_id( thisCode,
                                            sprintf( "$%d", {k} ),
                                            sprintf( "s[%d]", {k} ) )
                end for
        
                -- header    
                if j = Pre then
                    printf( outFile, "procedure ox_pre_%d( sequence s )\n", {i} )
                else
                    printf( outFile, "function ox_rule_%d( sequence s )\n", {i} )
                end if
                
                -- rule description
                puts( outFile, "    -- " )
                writeRule( outFile, i )
                
                -- the code
                puts( outFile, thisCode & "\n" )
        
                -- check for no return value
                if  j = Post
                and not match( "return ", thisCode ) then
                    -- default value, if none supplied
                    puts( outFile, "    -- default return value\n" )
                    puts( outFile, "    return 0\n" )
                end if
                
                
                -- close routine
                if j = Pre then
                    puts( outFile, "end procedure\n\n" )
                else
                    puts( outFile, "end function\n\n" )
                end if
                
            end if
        
        end for
    
    end for
end procedure

-----------------------------------------------------------------------------
function generateRange( sequence s )
    -- return the range of chars indicated by the string
    -- for example, "0-2" will return "012"
    integer i, startChar, endChar
    sequence chars

    chars = ""
    i = 1
    while i <= length(s)  do
        -- get first char
        startChar = s[i]
        if i < length(s)
        and s[i+1] = '-'
        and i+2 <= length( s ) then

            -- build range            
            endChar = s[i+2]
            for char = startChar to endChar do
                chars &= char
            end for       
            
            -- move skip past range
            i += 2                 
            
        else
            -- add to chars
            chars &= startChar
            
        end if
        
        -- move ahead
        i += 1
        
    end while   
    
    return chars
    
end function  

-----------------------------------------------------------------------------
function quote( sequence s )
    -- build quoted string
    sequence string
    string = ""
    for i = 1 to length( s ) do
        -- quote?
        if s[i] = '"' then
            -- embed 
            string &= "\\\""
        elsif s[i] = '\\' then
            -- embed
            string &= "\\\\"
        else
            -- no special processing
            string &= s[i]
        end if
    end for        
    return "\"" & string & "\""
end function


-----------------------------------------------------------------------------
procedure parseOxFile( sequence fName )
    -- parse an ox file   
    integer section
    sequence word, first, string

    -- set the file name (for debugger)
    fileName = fName
    
    -- open the file
    inFile = open( fileName, "r" )
    if inFile = -1 then
        printf( 1, "Unable to open %s\n", {fileName} )
        abort(0)
    end if
    
    -- clear flags
    inCode = False
    lineNum = 0
    section = 0
    ignoreCase   = False
    
    -- clear accumulators
    grammar = {}
    code = {}
    tokenList = {}
    ruleList = {}
    keywordList = {}
    syntaxList = {}
    fileText = {}
    
    while not eof( inFile ) do

        -- get text
        lineText = line_input( inFile )
        fileText &= {lineText}
        lineNum = length( fileText )

        if match( "--", lineText ) = 1
        or match( "--", lineText ) = 1 then
            -- ignore
            word = {}
        else
            -- convert to words
            word = string_to_words( lineText )
        end if

        
        if length( word ) then

            -- get first word
            first = word[1]
        
            if first[1] = '[' then
                -- section

                if equal( "[Syntax]", first ) then
                    -- start of syntax section
                    section = SyntaxSection
                           
                elsif equal( "[Grammar]", first ) then
                    section = GrammarSection
                    
                else
                    fatal( "Unknown section: %s", {first} )
                    
                end if

            elsif section = 0 then
                fatal( "Section has not been declared", {} )
            
            elsif section = SyntaxSection then
            
                -- check syntax
                if length( word ) < 2 then
                    fatal( "Expected <attribute> = <value>", {} )                    
                elsif not find( word[1], SyntaxVars ) then
                    fatal( "Unknown syntax attribute %s", {word[1]} )
                elsif not equal( word[2], "=" ) then
                    fatal( "Expected '=', not '%s')", {word[2]} )
                end if

                -- allow blank
                if length( word ) = 2 then
                    word &= {""}
                end if

                -- need to convert to numeric value?
                if equal( "IgnoreCase", first )
                or equal( "AbortOnError", first ) then

                    -- convert to integer
                    if equal( word[3], "Yes" ) then
                        string = "1"     
                        -- need to set internal flag?
                        if equal( "IgnoreCase", first ) then
                            ignoreCase = True
                        end if
                    else
                        string = "0"
                    end if 
                    

                elsif equal( "StringStart", first )
                or    equal( "StringEnd", first )
                or    equal( "StringEsc", first )
                or    equal( "CharStart", first )
                or    equal( "CharEnd", first )
                or    equal( "CharEsc", first ) then
                    -- check length
                    if length( word ) = 3
                    and length( word[3] ) != 1 then
                        fatal( "%s can only be 1 character wide", {first} )
                    end if
                              
                    -- quote it
                    string = quote( word[3] )
                              
                -- generate range?                    
                elsif equal( first, "InitKeyWordChars" )
                or    equal( first, "KeyWordChars" ) then
                    string = generateRange( word[3] )
                    string = quote( string )

                -- quote                    
                else
                    string = quote( word[3] )
                    
                end if
                    
                -- add to syntax list
                syntaxList &= {{ word[1], string }}
                

            elsif section = GrammarSection then

                -- parsing code            
                if inCode then
                    parseCode( lineText )

                -- start of code                    
                elsif equal( first, "{:" )
                or    equal( first, "<:" ) then
                    parseCode( lineText )

                -- new rule, same head                     
                elsif equal( word[1], "|" ) then
                    -- new grammar rule, same head as prior
                    parseRule( word )
                     
                -- new rule
                elsif length( word ) > 1
                and equal( word[2], "=" ) then
                    parseRule( word )
                
                else
                    -- abort with error
                    fatal( "Unknown grammar directive %s", {word[1]} )
                    
                end if
                
            end if
            
        end if
        
    end while

    close( inFile )
    
end procedure


-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- The Rule Generation Section
-----------------------------------------------------------------------------

-----------------------------------------------------------------------------
procedure ruleConflict( integer index, sequence s1, sequence s2,
                            sequence why )
                            
    -- two rules are in conflict.
    integer rule1, rule2

    -- get rule
    s1 = s1[1]
    s2 = s2[1]

    -- get index from end of list    
    rule1 = s1[length(s1)]
    rule2 = s2[length(s2)]

    -- display an error
    printf( 1, "Rule conflict on token #%d\n", {index} )
    puts( 1, why & "\n" )
    
    -- write the conflicting rules
    writeRule( 1, rule1 )
    puts( 1, "\n" )
    writeRule( 1, rule2 )
    puts( 1, "\n" )

    -- abort program
    abort(0)
    
end procedure


-----------------------------------------------------------------------------
procedure writeCallToCode( sequence rule, integer depth )
    -- display code
    integer index
    
    -- index is stored at end of rule
    index = rule[length(rule)]
    
    -- rule code
    tab(depth)
    
    -- is there code?
    if length( code[index][Post] ) then
        -- call rule
        --printf( outFile, "return ox_rule_%d( s )\n", {index} )
        printf( outFile, "{ OxCall, routine_id(\"ox_rule_%d\") } },\n", {index} )
    else
        -- return default value
        --printf( outFile, "return 0\n", {index} )
        puts( outFile, "{ OxTerminal } },\n" )
    end if
    
end procedure


-----------------------------------------------------------------------------
function sortGroups( sequence group, sequence head, integer depth )
    
    -- group contains rules grouped by common heads. head contains
    -- the head token common to the group. depth contains the index
    -- of the rule being examined.
    --
    -- each group of rules is ordered so that rules with tokens at
    -- the head are placed first, followed by rules with rules, and
    -- finally rules with no heads.
    --
    -- an error is generated for the following conditions:
    --
    -- 1. There are different head rules. For example, a rule might
    --    be written:
    --
    --      RULE:   foo BAR
    --          |   foo BAZZ
    --
    --    In this case, the parser will be unable to decide if the first
    --    or second rule is appropriate when depth=2.
    --
    -- 2. There is a terminal rule, as well as a rule at the same token.
    --    For example:
    --
    --      RULE:   foo BAR
    --          |   foo
    --
    --    In this case, the parser is again unable to decide which is the
    --    proper rule to select.
    --
    -- 3. There are two terminal rules of the same depth. For example:
    --
    --      RULE:   foo
    --          |   foo
    --
    --    Since these rules are identical, the parser is unable to decide
    --    which is correct.
    
    integer preRule, priorRule, priorEmpty, conflict
    sequence new, token, rule

    -- clear the new groups
    new = {}

    -- look for "PRE" conflicting with other rules
    if length( head ) > 1 then
        preRule = find( "PRE", head )
        if preRule then
            -- pick a rule to conflict
            if preRule = 1 then
                conflict = 2
            else
                conflict = 1
            end if
                  
            -- conflict: rule with PRE node
            ruleConflict( depth, group[preRule], group[conflict],
                "$pre code conflicts with other rules" )
                  
        end if
    end if
    
    -- first, copy the tokens...
    for i = 1 to length(head) do
        -- non-empty?
        if sequence( head[i] ) then
            -- get the head token
            token = head[i]
            -- token?
            if tokenType( token, "TOK" ) then
                new = append( new, group[i] )
            end if
        end if
    end for 
    
    -- then the rules
    rule = ""      
    priorRule = 0
    for i = 1 to length(head) do
        -- non-empty?
        if sequence( head[i] ) then
            -- get the head token
            token = head[i]
            -- rule?
            if tokenType( token, "RUL" ) then
                -- add to list
                new = append( new, group[i] )
                
                -- is this different that a prior rule *and* not the head?
                if depth != 0
                and length( rule )
                and not equal( rule, token ) then
                    -- conflict: two rules on same node
                    ruleConflict( depth, group[i], group[priorRule],
                        "Can't choose between two rules" )
                else
                    -- set as rule
                    rule = token
                    priorRule = i
                end if
                
            end if
        end if
    end for

    -- then the empty rules...
    priorEmpty = 0
    for i = 1 to length(head) do
        -- empty?
        if integer( head[i] ) then
            -- add to list
            new = append( new, group[i] )
            
            -- is there a rule at this node?
            if depth and priorRule then
                -- can't decide between rule and empty
                ruleConflict( depth, group[priorRule], group[i],
                    "Can't decide between rule and terminal" )
                
            end if
            
            -- was there another empty node?
            if depth and priorRule then
                -- can't decide between two empty rules
                ruleConflict( depth, group[priorEmpty], group[i],
                    "Can't decide between two terminals" )
            else
                -- set as prior empty rule
                priorEmpty = i                
            end if
            
        end if
    end for

    return new
    
end function

-----------------------------------------------------------------------------
function groupRules( sequence rules, integer depth )

    integer flag
    sequence group, head

    -- group all types into similar heads
    group = {}              
    head = {}

    -- look at each rule
    for i = 1 to length(rules) do

        -- look for a rule bucket with the same head rule
        flag = False
        for j = 1 to length( group ) do
            -- same head?
            if equal( rules[i][1], head[j] ) then
                -- add to bucket
                group[j] = append( group[j], rules[i] )
                -- set flag, new bucket not needed
                flag = True
                -- leave the loop
                exit
            end if
        end for
        -- no match?
        if flag = False then
            -- put in new bucket
            group = append( group, {rules[i]} )
            -- save head
            head = append(head, rules[i][1] )
        end if
    end for

    return sortGroups( group, head, depth )

end function


-----------------------------------------------------------------------------
function removeHeads( sequence s )

    -- remove the heads from all the rule

    sequence rule

    for i = 1 to length( s ) do
        -- get the rule
        rule = s[i]    
        
        -- remove the head
        rule = rule[2..length(rule)]

        -- replace
        s[i] = rule
        
    end for 

    return s
    
end function

-----------------------------------------------------------------------------
procedure writeTest( sequence token, integer ruleNum )

    -- write test based on object being parsed
    -- ruleNum is only used for "PRE" rules
    -- testNum is used to suppress OxExpect if it's the first rule

    -- rule?
    if tokenType( token, "RUL" ) then
        -- remove rule flag
        token = stripToken( token )
        
        -- call oxParse to validate
        --printf( outFile, "s &= {oxParse_( \"%s\" )}\n", {token} )
        printf( outFile, "{ OxParse, \"%s\" },\n", {token} )

    -- rule?
    elsif tokenType( token, "PRE" ) then   
        -- call pre routine
        --printf( outFile, "ox_pre_%d(s)\n", {ruleNum} )
        printf( outFile, "{ OxPre, routine_id(\"ox_pre_%d\") },\n", {ruleNum} )

    -- literal or token
    else
       
        -- remove literal flag
        token = stripToken( token )  
        
        -- make case insensitive        
        if ignoreCase then
            token = upper( token )
        end if

        -- call oxTokenGet to validate
        --printf( outFile, "s &= {oxExpect(\"%s\")}\n", {token} )
        printf( outFile, "{ OxExpect, \"%s\"},\n", {token} )

                
    end if

end procedure

-----------------------------------------------------------------------------
procedure writeHeadTest( sequence group, integer index, integer depth )

    -- write a test to differentiate one rule from another
    sequence head

    -- write an if statement for the head of the rule
    tab(depth)

    -- get the head          
    if length( group[index][1] ) = 1 then
        -- empty
        head = ""
    else
        -- get the head of the rule        
        head = group[index][1][1]
    end if

    -- empty head?        
    if length( head ) = 0 then
        -- terminal rule. since there is only one terminal rule allowed
        -- per group, and terminal rules are placed at the end of the
        -- list, all rules have been matched. assume the result is true.
        puts( outFile, "{ OxTrue,\n" )

    -- rule?
    elsif tokenType( head, "RUL" ) then
        -- there can only be on rule per group (at the same level), and
        -- rules and terminal rules are mutually exclusive. assume the
        -- test result to be true.
        --puts( outFile, "1" )
        puts( outFile, "{ OxTrue,\n" )
    
    -- token
    else            
        -- strip type off token
        head = stripToken( head )

        -- make case insensitive        
        if ignoreCase then
            head = upper( head )
        end if
    
        -- match literal          
        printf( outFile, "{ OxToken, \"%s\",\n", {head} )
        
    end if

end procedure


-----------------------------------------------------------------------------
procedure generateTests( sequence group, integer index, integer depth )

    -- this routine is passed a list of rules, the current index into
    -- the rules, and the depth of recursion. it groups the rules by
    -- common heads, and then removes the heads from the rules until
    -- the lists are unique.
    
    integer thisIndex, ruleNum
    sequence thisGroup, rules, head

    -- go through each group
    for i = 1 to length(group) do

        -- write a test for this element
        writeHeadTest( group, i, depth )

        -- only item in group?
        if length( group[i] ) = 1 then

            -- get the rules
            rules = group[i][1]
            
            -- get the rule number
            ruleNum = rules[length(rules)]

            -- decode the remainder of the rule
            for j = 1 to length( rules )-1 do  
                                                
                -- write the test
                tab(depth+1)
                writeTest( rules[j], ruleNum )

            end for

            -- get the rule
            tab(depth+1)
            puts( outFile, "-- " )
            writeRule( outFile, rules[length(rules)]  )
            puts( outFile, "\n" )

            -- call the code
            writeCallToCode( rules, depth+1 )
            
        else

            -- get the group
            thisGroup = group[i]

            -- set the index
            thisIndex = index

            -- until there is a difference in the group
            while length( thisGroup ) do

                -- get the head of the group
                head = thisGroup[1][1]
                
                -- write test for the head
                tab(depth+1)    
                writeTest( head, 0 )
                
                -- remove the heads
                thisGroup = removeHeads( thisGroup )

                -- regroup
                thisGroup = groupRules( thisGroup, depth )

                -- different heads if regrouped?
                if length( thisGroup ) != 1 then
                
                    -- recurse
                    generateTests( thisGroup, thisIndex+1, depth+1 )
                    
                    -- leave loop
                    exit
                    
                else    
                    -- get single group
                    thisGroup = thisGroup[1]

                    -- only item in group?
                    if length( thisGroup ) = 1 then
                        -- finish rule
                        generateTests( thisGroup, thisIndex+1, depth+1 )
                        
                        -- leave loop
                        exit
                        
                    end if
                
                end if

                -- increment the index
                thisIndex += 1                   
                   
            end while
        
            
        end if
        
    end for

    -- generate code
    tab(depth)                     
    puts( outFile, "{ OxFailed } },\n" )
    
end procedure


-----------------------------------------------------------------------------
procedure verifyGrammar( sequence rules )
    -- verify that all the tokens that refer to rules actually
    -- are defined as rules     
    integer ruleNum
    sequence list, rule, token

    -- gather a list of all rule names
    list = {}
    for i = 1 to length( rules ) do
        token = stripToken( rules[i][1] )
        if not find( token, list ) then
            list &= {token}
        end if
    end for
    
    -- look through the items
    for i = 1 to length( rules ) do
        rule = rules[i]
        for j = 2 to length( rule )-1 do
            -- get a token
            token = rule[j]
            -- is it a rule?
            if tokenType( token, "RUL" ) then
                -- strip the type off
                token = stripToken( token )
                -- not in the rules list?
                if not find( token, list ) then
                    -- get rule number at end of rule
                    ruleNum = rule[length(rule)]
                    -- error, no line number       
                    puts( 1, "Error found in rule\n" )
                    writeRule( 1, ruleNum )           
                    puts( 1, "\n" )
                    printf( 1, "Rule %s not found\n", {token} )
                    abort(0)
                end if
            end if
        end for
    end for
    
end procedure


-----------------------------------------------------------------------------
procedure writeGrammar( sequence rules )

    -- This is passed the grammar rules. It generates the
    -- top level tests, and then calls generateRules to
    -- evaluate each set of grammar rules.

    sequence group, rule

    -- assign to oxParseRules
    puts( outFile, "oxParseRules = {\n" )

    -- group the rules
    rules = groupRules( rules, 0 )

    -- go through each group
    for i = 1 to length(rules) do

        -- get the rule token (first token of first rule)
        rule = rules[i][1][1]

        -- strip type off token
        rule = stripToken( rule )

        -- the test
        --printf( outFile, "equal( rule, \"%s\" ) then\n", {rule} )
        printf( outFile, "{ OxRule, \"%s\",\n", {rule} )

        -- remove the heads
        group = removeHeads( rules[i] )

        -- regroup
        group = groupRules( group, 1 )

        -- generate the rule tests
        generateTests( group, 1, 1 )
        
    end for

    -- generate error code
    puts( outFile, "{ OxFailed } }\n" )
                               
end procedure



-----------------------------------------------------------------------------
procedure main( sequence inName )

    integer at
    sequence grouped, list, outName

    -- remove extention          
    outName = "ox_" & inName
    at = find( '.', outName )         
    if at then
        outName = outName[1..at-1]
    end if
    outName = outName & ".e"

    -- open the output file
    outFile = open( outName, "w" )

    -- read the input file
    parseOxFile( inName )
    
    -- file header
    printf( outFile, "-- Parser Generated by %s %s\n", {AppName, AppVersion} )
    printf( outFile, "-- %s %s\n\n", {AppName, AppCopyright})
    puts( outFile, "without warning\n" )
    
    -- include file                   
    puts( outFile, "-- the lexer\n" )
    puts( outFile, "include ox_lex.e\n\n" )

    -- tokens
    puts( outFile, "-- Syntax\n" )     
    for i = 1 to length( syntaxList ) do                     
        printf( outFile, "ox%s = %s\n", syntaxList[i] )
    end for
    puts( outFile, "\n" )

    -- keywords                      
    keywordList = sort( keywordList )
    puts( outFile, "-- Keywords\n" )     
    -- group the keyword list
    grouped = group( keywordList )
    for i = 1 to length( grouped ) do
        -- sort
        list = sort( grouped[i] )        
        -- in reverse (length) order
        for j = length( list ) to 1 by -1 do
            printf( outFile, "oxKeyword &= {\"%s\"}\n", {list[j]} )
        end for
    end for
    puts( outFile, "\n" )

    -- tokens
    puts( outFile, "-- Tokens\n" )     
    list = sort( tokenList )
    for i = 1 to length( list ) do                     
        if not find( list[i], keywordList ) then
            printf( outFile, "oxLiteral &= {\"%s\"}\n", {list[i]} )
        end if
    end for
    puts( outFile, "\n" )
    
    -- add rule numbers
    for i = 1 to length( grammar ) do
        grammar[i] = append( grammar[i], i )
    end for
    
    -- generate rule code
    writeRuleCode()

    -- verify the tokens in the grammar
    verifyGrammar( grammar )
    
    -- generate the grammar
    writeGrammar( grammar )
        
    close( outFile )

    printf( 1, "Created file %s\n", {outName} )

end procedure

sequence cmd
cmd = command_line()
if length( cmd ) < 3 then
    -- prompt the user
    cmd = prompt_string( "Ox file to convert? " )
    if length( cmd ) then
        main( cmd )
    end if
else
    main(cmd[3])
end if

