-- parse.e
-- parse routine for ox


include get.e
include string.e
include wildcard.e

with trace

constant
    False        = 0,
    True         = not False,
    DecimalChars =  "1234567890",
    HexChars     = DecimalChars & "abcdefABCDEF",
    WhiteSpace   = "\n\t"
    

integer 
    oxLineNum,          -- current line number, index to oxText
    oxTokenIndex,       -- current token, index to oxTokens
    oxEvalWordId,       -- routine id of oxEvaluateWord routine
    oxInComment         -- true if parser is inside a comment

sequence
    oxFileName,         -- name of file oxText came from
    oxStrings,           -- sequences of text to evaluate
    oxString,           -- current string being read
    oxTokens,           -- list of tokens parsed from oxText
    oxPrior             -- stack of prior file being read

global object
    oxTokenValue    -- value of current oxToken

global sequence
    oxToken,
    oxTokenType     -- here for debugging

global sequence 
    oxKeyword,      -- list of keywords
    oxLiteral,      -- list of literals
    oxWhiteSpace,   -- list of whitespace characters
    oxResume        -- grammar rule to resume parsing on    

global integer
    oxError         -- true if encountered an error

    
    oxKeyword       = {}
    oxLiteral       = {}
    oxWhiteSpace    = {}
    oxResume        = ""

global sequence
    oxInitKeyWordChars,
    oxKeyWordChars,
    oxCommentStart,
    oxCommentEnd,
    oxSingleComment,
    oxStringStart,
    oxStringEnd,
    oxStringEsc,
    oxCharStart,
    oxCharEnd,
    oxCharEsc,
    oxHexStart

global integer    
    oxAbortOnError,
    oxIgnoreCase,
    oxParseRoutine      -- initialized by the parser
    
-- defaults
    oxAbortOnError      = 1
    oxIgnoreCase        = 0
    oxInitKeyWordChars  = "qwertyuiopasdfghjklzxcvbnm" &
                          "QWERTYUIOPASDFGHJKLZXCVBNM"
    oxKeyWordChars      = oxInitKeyWordChars & DecimalChars & "_"
    oxCommentStart      = "/*"
    oxCommentEnd        = "*/"
    oxSingleComment     = "//"
    oxStringStart       = "\""
    oxStringEnd         = "\""
    oxStringEsc         = "\\"
    oxCharStart         = "'"
    oxCharEnd           = "'"
    oxCharEsc           = "\\"
    oxHexStart          = "0x"
    


-----------------------------------------------------------------------------
procedure oxEvaluateWord( sequence word )
    sequence result
    result = call_func( oxEvalWordId, {word} )
    oxToken = word
    oxTokenType = result[1]
    oxTokenValue = result[2]
end procedure


-----------------------------------------------------------------------------
global procedure oxSetEvalWord( integer id )
    oxEvalWordId = id
end procedure

-----------------------------------------------------------------------------
function oxAddEscChars( sequence s )
    -- add escape characters to string
    integer esc, char, flag
    sequence new
    new = {}   
    esc = oxStringEsc[1]   
    flag = 0
    for i = 1 to length( s ) do
        if s[i] = esc then
            -- flag next char as escaped
            flag = i+1
        elsif i = flag then
            -- convert to special char
            if s[i] = 'n' then char = '\n'
            elsif s[i] = 'r' then char = '\r'
            elsif s[i] = 't' then char = '\t'
            elsif s[i] = '\\' then char = '\\'
            elsif s[i] = '\"' then char = '\"'
            elsif s[i] = '\'' then char = '\''
            else char = s[i]
            end if
            new &= char
            flag = 0
        else
            new &= s[i]
        end if
    end for   
    return new
end function

-----------------------------------------------------------------------------
function oxDefaultEvalWord( sequence word )

    -- evaluate a word, and sets the token and return values
    object val

    if find( word, oxKeyword ) then
        -- return as a literal
        return { word, word }

    -- strings start with quote characters
    elsif match( oxStringStart, word ) = 1 then
        -- remove quotes
        word = word[2..length(word)-1]
        -- handle escape chars
        word = oxAddEscChars( word )
        return { "STRING", word }

    -- character constant?
    elsif word[1] = oxCharStart[1] then
        -- remove quotes
        word = word[2..length(word)-1]
        -- handle escape chars
        word = oxAddEscChars( word )
        return { "CHAR", word[1] }
        
    -- decimal number?
    elsif find( word[1], DecimalChars & '.' ) then
        val = value( word )
        return { "NUMBER", val[2] }

    -- hexidecimal number?
    elsif match( oxHexStart, word ) = 1 then           
        -- remove tag
        word = word[length(oxHexStart)+1..length(word)]
        
        -- prefix with '#' and set to upper case
        word = "#" & upper( word )
        
        -- convert
        val = value( word )
        return { "NUMBER", val[2] }
    
    else
        -- unknown
        return { "IDENTIFIER", word }
        
    end if
    
end function

-----------------------------------------------------------------------------
-- set default value
oxSetEvalWord( routine_id( "oxDefaultEvalWord" ) )


-----------------------------------------------------------------------------
global procedure oxParseError( sequence message, sequence args )

    -- skip if error was already displayed;
    -- this avoids cascading errors
    if oxError then
        return
    end if

    -- show the errant line
    puts( 1, oxString & "\n" )

    -- convert special characters
    for i = 1 to length( args ) do
        if equal( args[i], "\n" ) then
            args[i] = "<end of line>"
        elsif equal( args[i], "\t" ) then
            args[i] = "<tab>"
        end if
    end for

    -- print the parse error information, and close the program
    printf( 1, message & "\n", args )

-- Perhaps it should unwind down the oxPrior stack as well
    
    -- was a file open?
    if oxLineNum then          
        if length(oxFileName) then
            -- show file name and line number
            printf( 1, "File: %s, Line Number %d\n", {oxFileName, oxLineNum} )
        end if
    end if
    
    -- shut down?
    if oxAbortOnError then
        abort(0)    
    end if

    -- set error flag
    oxError = True

end procedure

-----------------------------------------------------------------------------
procedure oxSaveState( sequence fileName )

    -- save the current state of the input text

    -- store the current state
    oxPrior &= {{   oxLineNum,
                    oxTokenIndex,
                    oxFileName,
                    oxStrings,
                    oxString,
                    oxTokens,
                    oxToken,
                    oxTokenType,
                    oxTokenValue  }}


    -- clear state
    oxLineNum       = 0
    oxTokenIndex    = 0
    oxFileName      = fileName
    oxStrings       = {}
    oxString        = ""
    oxTokens        = {}
    oxToken         = ""
    oxTokenType     = ""
    oxTokenValue    = 0
    

    
end procedure

-----------------------------------------------------------------------------
global procedure oxLoadBuffer( sequence bufferName, sequence buffer )

    -- set ox to parse the buffer
    
    -- save the current state
    oxSaveState( bufferName )

    -- set the buffer as input
    oxStrings = buffer

end procedure
-----------------------------------------------------------------------------
global procedure oxLoadFile( sequence fileName )

    -- read a file into oxText, set oxFileName
    -- if an error is encountered, prints a message and aborts.
    -- the prior file that was being processed is stored on
    -- the stack.

    integer handle
    object data

    -- attempt to open the file
    handle = open( fileName, "r" )
    if handle = -1 then
        -- try to recover
        oxParseError( "Can't open file '%s'.", {fileName} )
        data = {}
        oxStrings = {}
        oxSaveState( fileName )
        return
    end if
    
    -- save the current state
    oxSaveState( fileName )

    -- read until end of file
    while 1 do

        -- read a line
        data = gets( handle )
                    
        -- eof?
        if integer( data ) then
            
            -- leave loop
            exit

        else            
            -- accumulate text
            oxStrings &= {data}  
            
        end if
        
    end while

    -- close the file
    close( handle )
    
end procedure


-----------------------------------------------------------------------------
global procedure oxUnloadFile()

    -- close the current file. if a prior file was open, 
    -- the state of that file is restored.

    sequence prior

    if length( oxPrior ) then

        -- get top stack item
        prior = oxPrior[ length(oxPrior) ]
        
        -- pop from stack
        oxPrior = oxPrior[1..length(oxPrior)-1 ]
        
        -- restore prior state
        oxLineNum       = prior[1]
        oxTokenIndex    = prior[2]
        oxFileName      = prior[3]
        oxStrings       = prior[4]
        oxString        = prior[5]
        oxTokens        = prior[6]
        oxToken         = prior[7]
        oxTokenType     = prior[8]
        oxTokenValue    = prior[9]
        
    end if
    
end procedure


-----------------------------------------------------------------------------
global procedure oxUnloadBuffer()
    -- same as oxUnloadFile
    oxUnloadFile()
end procedure
                                                

-----------------------------------------------------------------------------
function oxExtractQuote( sequence string, sequence endOf, sequence escOf )
    -- returns quote from string, assumes to start from position 1
    -- if not found, issue an error
    for i = 2 to length( string ) do
        if string[i-1] != escOf[1]
        and string[i] = endOf[1] then
            return string[1..i]
        end if
    end for
    oxParseError( "End of line reached with no closing %s", {endOf} )
    return {}
end function

-----------------------------------------------------------------------------
function oxMatchWord( sequence word, sequence string )
    -- returns true if word is at the beginning of string
    -- the word must *not* be followed by a keyword char
    
    -- special char?
    if not find( string[1], oxKeyWordChars ) then
        if      length( word )
        and     length( string )
        and     length( word ) < length( string )
        and     match( word, string ) = 1
        then
            return True
        else
            return False
        end if
    else
        -- char, so must be seperated with delimiter
        if      length( word )
        and     length( string )
        and     length( word ) < length( string )
        and     match( word, string ) = 1
        and     not find( string[length(word)+1], oxKeyWordChars ) then
            return True
        else
            return False
        end if
    end if
end function

-----------------------------------------------------------------------------
function atLiteral( sequence s )
    -- return literal that lies at the given position
    integer flag
    flag = 0
    for i = 1 to length( oxKeyword ) do          
        if oxMatchWord( oxKeyword[i], s ) then
            return i
        end if
    end for
    return 0
end function


-----------------------------------------------------------------------------
function oxStringToWords( sequence string )
    
    -- convert a sequence of chars into words
    
    integer decimal, at
    sequence words, word

    words = {}
    word = ""
    string &= " "

    while length( string ) do
    
        -- white space?
        if find( string[1], oxWhiteSpace ) then
            -- noise
            word = " "

        -- string?          
        elsif length( oxStringStart )
        and   match( oxStringStart, string ) = 1 then
            word = oxExtractQuote( string, oxStringEnd, oxStringEsc )

        -- char?
        elsif length( oxCharStart )
        and   match( oxCharStart, string ) = 1 then
            word = oxExtractQuote( string, oxCharEnd, oxCharEsc )

            -- check length
            if length( word ) = 4
            and word[2] != oxCharEsc[1] then
                oxParseError( "Character constant %s too large", {word} )
                return word
            elsif length( word ) != 3 then
                oxParseError( "Character constant %s too large", {word} )
                return word
            end if

        -- start of comment?
        elsif length( oxCommentStart )
        and   match( oxCommentStart, string ) = 1 then
            word = oxCommentStart

            -- in comment?
            if not oxInComment then
                -- set flag
                oxInComment = True
            end if
            
        -- end of comment?     
        elsif length( oxCommentEnd )
        and   match( oxCommentEnd, string ) = 1 then
            word = oxCommentEnd

        -- line comment?
        elsif not oxInComment
        and   length( oxSingleComment )
        and   match( oxSingleComment, string ) = 1 then
            -- all done
            exit

        else
            -- literal?
            at = atLiteral( string )
            if at then
                word = oxKeyword[at]
                
            -- not a literal?
            else
                           
                -- start of keyword?
                if find( string[1], oxInitKeyWordChars ) then
                    -- build keyword
                    word = {string[1]}
                    for i = 2 to length( string ) do
                        if find( string[i], oxKeyWordChars ) then
                            word &= string[i]
                        else
                            -- leave loop
                            exit
                        end if
                    end for
            
                -- decimal number?
                elsif find( string[1], DecimalChars & '.' ) then
                    -- parse number
                    word = {}                      
                    decimal = 0
                    for i = 1 to length( string ) do
                        -- digit?
                        if find( string[i], DecimalChars ) then
                            word &= string[i]
                        
                        -- decimal?
                        elsif string[i] = '.' then
                            -- already encountered?
                            if decimal then
                                -- end of number
                                exit
                            else          
                                -- part of literal? (ex: ".." )
                                at = atLiteral( string[i..length(string)] )
                                if at and not equal( oxKeyword[at], "." )
                                then
                                    exit
                                else
                                    -- add decimal point, set flag
                                    word &= string[i]
                                    decimal = 1
                                end if
                            end if
                        else
                            exit
                        end if
                    end for
        
        
                -- hexidecimal number?                
                elsif length( oxHexStart )
                and   match( oxHexStart, string ) = 1 then
                    -- parse hex number
                    word = oxHexStart
                    for i = length(oxHexStart)+1 to length( string ) do
                        if find( string[i], HexChars ) then
                            word &= string[i]
                        else
                            exit
                        end if
                    end for     
            
            
                else
                    -- treat char as word
                    word = {string[1]}
                    
                end if
            end if
        end if

        -- remove word from string
        string = string[length(word)+1..length(string)]

        -- accumulate?
        if not oxInComment
        and not equal( word, " " ) then
            words &= {word}
        end if


        -- turn off comment flag?
        if equal( word, oxCommentEnd ) then
            oxInComment = False
        end if
                    
            
    end while
    
    return words

end function

-----------------------------------------------------------------------------
procedure oxGetNextLine()

    -- convert the next line of text in oxText into
    -- a string of tokens

    -- end of document?
    if oxLineNum >= length( oxStrings ) then
        -- end of file is current line
        oxString = "<EOF>"
        
        -- set as single token returned
        oxTokens = { oxString }
    
    else
        -- increment line number
        oxLineNum += 1
        
        -- set current line
        oxString = oxStrings[oxLineNum]
        
        -- parse current line
        oxTokens = oxStringToWords( oxString )

    end if

    -- debugging
    -- puts( 1, oxString )
    
end procedure


-----------------------------------------------------------------------------
procedure oxMoveAhead()
    -- get the next token from the buffer
    oxTokenIndex += 1
    oxToken = ""
end procedure


-----------------------------------------------------------------------------
function oxCurrentToken()

    -- need to parse a new line of tokens?
    if length( oxTokens ) < oxTokenIndex
    or oxTokenIndex = 0 then
        -- clear token list and index to current token
        oxTokens = {}      
        oxTokenIndex = 1   
        
        -- read until tokens are accumulated
        while length(oxTokens) = 0 do
            -- get the next line from the strings, and tokenize
            oxGetNextLine()
        end while
    end if    
    
    -- no current token?
    if length( oxToken ) = 0 then
        oxEvaluateWord( oxTokens[oxTokenIndex] )
    end if

    return oxTokenType
    
end function


-----------------------------------------------------------------------------
global function oxTokenGet()
    sequence word
    word = oxCurrentToken()
    oxMoveAhead()
    return word
end function


-----------------------------------------------------------------------------
global function oxTokenIs( sequence word )
    if oxIgnoreCase then
        return equal( upper(oxCurrentToken()), word )
    else
        return equal( oxCurrentToken(), word )
    end if
end function

-----------------------------------------------------------------------------
global function oxEatToken( sequence word )
    if oxTokenIs( word ) then
        oxMoveAhead()
        return True
    else
        return False        
    end if
end function

-----------------------------------------------------------------------------
global function oxExpect( sequence word )
    -- the word must match the token.
    -- if it does, returns the token's value.
    -- if not, generates an error
    if oxEatToken( word ) then
        return oxTokenValue
    else
        oxParseError( "Expected \"%s\", not \"%s\"", {word, oxToken} )
        return False
    end if
end function


-----------------------------------------------------------------------------
global function oxParse( sequence rule )
    -- clear error flag and call the parser
    oxError = 0
    return call_func( oxParseRoutine, { rule, 0 } )
end function

-----------------------------------------------------------------------------
global procedure oxInit()              

    -- called before running the parser

    -- initialize state
    oxLineNum       = 0
    oxTokenIndex    = 0
    oxFileName      = ""
    oxStrings       = {}
    oxString        = ""
    oxTokens        = {}
    oxToken         = ""
    oxTokenType     = ""
    oxPrior         = {}
    oxTokenValue    = 0
    oxInComment     = False

    -- whatever whitespace chars are not literals are added
    -- to the list of whitespace characters
    oxWhiteSpace = " "
    for i = 1 to length( WhiteSpace ) do     
        -- not literal?
        if not find( {WhiteSpace[i]}, oxKeyword ) then
            oxWhiteSpace &= WhiteSpace[i]
        end if
    end for
    
end procedure

