-- parse.e
-- parse routine for ox


include get.e
include sort.e
include string.e
include wildcard.e

with trace
without warning

-- parse tree tokens
global constant
    OxCall      = 0,        -- end of parse, call routine
    OxExpect    = 1,        -- token must match
    OxFailed    = 2,        -- parse failed
    OxParse     = 3,        -- call parsing rule
    OxPre       = 4,        -- pre action embedded in parse
    OxRule      = 5,        -- branch if matches rule
    OxToken     = 6,        -- branch if matches token
    OxTrue      = 7,        -- always branch
    OxTerminal  = 8         -- end of parse, but no action

constant
    False        = 0,
    True         = not False,
    DecimalChars =  "1234567890",
    HexChars     = DecimalChars & "abcdefABCDEF",
    WhiteSpace   = "\n\t\r"
    

global integer 
    oxError,            -- true if encountered an error
    oxEvalWordId,       -- routine id of oxEvaluateWord routine
    oxInComment,        -- true if parser is inside a comment
    oxLineId,           -- line first token was found on
    oxLineNum,          -- current line number, index to oxText
    oxTokenIndex        -- current token, index to oxTokens

global sequence
    oxFileName,         -- name of file oxText came from
    oxKeyword,          -- list of keywords
    oxLiteral,          -- list of literals
    oxParseRules,       -- parse rules supplied by the grammar
    oxPrior,            -- stack of prior file being read
    oxResume,           -- grammar rule to resume parsing on    
    oxString,           -- current string being read
    oxStrings,          -- sequences of text to evaluate
    oxToken,            -- text of token
    oxTokenType,        -- type of token
    oxTokens,           -- list of tokens parsed from oxText
    oxWhiteSpace        -- list of whitespace characters

global object
    oxTokenValue    -- value of current oxToken
    
-----------------------------------------------------------------------------
-- these are needed before oxInit() is called
    oxKeyword       = {}
    oxLiteral       = {}
    oxWhiteSpace    = {}
    oxResume        = ""

-----------------------------------------------------------------------------
-- attributes of the grammar
global integer
    oxAbortOnError,     
    oxIgnoreCase

global sequence
    oxInitKeyWordChars,
    oxKeyWordChars,
    oxCommentStart,
    oxCommentEnd,
    oxSingleComment,
    oxSingleComment2,
    oxStringStart,
    oxStringEnd,
    oxStringEsc,
    oxCharStart,
    oxCharEnd,
    oxCharEsc,
    oxHexStart
    
-- defaults
    oxAbortOnError      = 1
    oxIgnoreCase        = 0
    oxInitKeyWordChars  = "qwertyuiopasdfghjklzxcvbnm" &
                          "QWERTYUIOPASDFGHJKLZXCVBNM"
    oxKeyWordChars      = oxInitKeyWordChars & DecimalChars & "_"
    oxCommentStart      = "/*"
    oxCommentEnd        = "*/"
    oxSingleComment     = "//"
    oxSingleComment2    = ""
    oxStringStart       = "\""
    oxStringEnd         = "\""
    oxStringEsc         = "\\"
    oxCharStart         = "'"
    oxCharEnd           = "'"
    oxCharEsc           = "\\"
    oxHexStart          = "0x"
    
-----------------------------------------------------------------------------
procedure oxEvaluateWord( sequence word )
    sequence result
    result = call_func( oxEvalWordId, {word} )
    oxToken = word
    oxTokenType = result[1]
    oxTokenValue = result[2]
end procedure


-----------------------------------------------------------------------------
global procedure oxSetEvalWord( integer id )
    oxEvalWordId = id
end procedure

-----------------------------------------------------------------------------
function oxAddEscChars( sequence s )
    -- add escape characters to string
    integer esc, char, flag
    sequence new
    
    new = {}   
    esc = oxStringEsc[1]   
    flag = 0
    for i = 1 to length( s ) do
        if i = flag then
            -- convert to special char
            if s[i] = 'n' then char = '\n'
            elsif s[i] = 'r' then char = '\r'
            elsif s[i] = 't' then char = '\t'
            elsif s[i] = '\\' then char = '\\'
            elsif s[i] = '\"' then char = '\"'
            elsif s[i] = '\'' then char = '\''
            else char = s[i]
            end if
            new &= char
            flag = 0
        elsif s[i] = esc then
            -- flag next char as escaped
            flag = i+1
        else
            new &= s[i]
        end if
    end for   
    return new
end function

-----------------------------------------------------------------------------
function oxDefaultEvalWord( sequence word )

    -- evaluate a word, and sets the token and return values
    object val

    if find( word, oxKeyword ) then
        -- return as a literal
        return { word, word }

    -- strings start with quote characters
    elsif match( oxStringStart, word ) = 1 then
        -- remove quotes
        word = word[2..length(word)-1]
        -- handle escape chars
        word = oxAddEscChars( word )
        return { "STR", word }

    -- character constant?
    elsif word[1] = oxCharStart[1] then
        -- remove quotes
        word = word[2..length(word)-1]
        -- handle escape chars
        word = oxAddEscChars( word )
        return { "CHAR", word[1] }
        
    -- decimal number?
    elsif find( word[1], DecimalChars ) then
        val = value( word )
        return { "NUM", val[2] }
                      
    -- decimal number starting with decimal?
    elsif length( word ) > 1
    and word[1] = '.'
    and find( word[2], DecimalChars ) then
        val = value( word )
        return { "NUM", val[2] }
                      
    -- hexidecimal number?
    elsif match( oxHexStart, word ) = 1 then           

        -- remove tag
        word = word[length(oxHexStart)+1..length(word)]
        
        -- prefix with '#' and set to upper case
        word = "#" & upper( word )
        
        -- convert
        val = value( word )
        return { "NUM", val[2] }
    
    else
        -- unknown
        return { "ID", word }
        
    end if
    
end function

-----------------------------------------------------------------------------
-- set default value
oxSetEvalWord( routine_id( "oxDefaultEvalWord" ) )


-----------------------------------------------------------------------------
global procedure oxParseError( sequence message, sequence args )

    -- skip if error was already displayed;
    -- this avoids cascading errors
    if oxError then
        return
    end if

    -- convert special characters
    for i = 1 to length( args ) do
        if equal( args[i], "\n" )
        or equal( args[i], "\r" ) then
            args[i] = "<end of line>"
            
        elsif equal( args[i], "\t" ) then
            args[i] = "<tab>"
            
        elsif equal( args[i], "<EOF>" ) then
            args[i] = "<end of file>"
            
        end if
    end for

    -- print the parse error information
    printf( 1, "\nError: " & message & "\n", args )

-- Perhaps it should unwind down the oxPrior stack as well
    
    -- was a file open?
    if oxLineNum then          
        if length(oxFileName) then
            -- show file name and line number
            printf( 1, "File: %s, Line Number %d\n", {oxFileName, oxLineNum} )
        end if
    end if

    -- show the errant line
    puts( 1, oxString & "\n" )
    
    -- shut down?
    if oxAbortOnError then
        abort(0)    
    end if

    -- set error flag
    oxError = True

end procedure



-----------------------------------------------------------------------------
procedure oxSaveState( sequence fileName )

    -- save the current state of the input text

    -- store the current state
    oxPrior &= {{   oxLineNum,
                    oxTokenIndex,
                    oxFileName,
                    oxStrings,
                    oxString,
                    oxTokens,
                    oxToken,
                    oxTokenType,
                    oxTokenValue  }}


    -- clear state
    oxLineNum       = 0
    oxTokenIndex    = 0
    oxFileName      = fileName
    oxStrings       = {}
    oxString        = ""
    oxTokens        = {}
    oxToken         = ""
    oxTokenType     = ""
    oxTokenValue    = 0
    

    
end procedure

-----------------------------------------------------------------------------
global procedure oxLoadBuffer( sequence bufferName, sequence buffer )

    -- set ox to parse the buffer
    
    -- save the current state
    oxSaveState( bufferName )

    -- set the buffer as input
    oxStrings = buffer

end procedure
-----------------------------------------------------------------------------
global procedure oxLoadFile( sequence fileName )

    -- read a file into oxText, set oxFileName
    -- if an error is encountered, prints a message and aborts.
    -- the prior file that was being processed is stored on
    -- the stack.

    integer handle
    object data

    -- attempt to open the file
    handle = open( fileName, "r" )
    if handle = -1 then
        -- try to recover
        oxParseError( "Can't open file '%s'.", {fileName} )
        data = {}
        oxStrings = {}
        oxSaveState( fileName )
        return
    end if
    
    -- save the current state
    oxSaveState( fileName )

    -- read until end of file
    while 1 do

        -- read a line
        data = gets( handle )
                    
        -- eof?
        if integer( data ) then
            
            -- leave loop
            exit

        else            
            -- accumulate text
            oxStrings &= {data}  
            
        end if
        
    end while

    -- close the file
    close( handle )
    
end procedure


-----------------------------------------------------------------------------
global procedure oxUnloadFile()

    -- close the current file. if a prior file was open, 
    -- the state of that file is restored.

    sequence prior

    if length( oxPrior ) then

        -- get top stack item
        prior = oxPrior[ length(oxPrior) ]
        
        -- pop from stack
        oxPrior = oxPrior[1..length(oxPrior)-1 ]
        
        -- restore prior state
        oxLineNum       = prior[1]
        oxTokenIndex    = prior[2]
        oxFileName      = prior[3]
        oxStrings       = prior[4]
        oxString        = prior[5]
        oxTokens        = prior[6]
        oxToken         = prior[7]
        oxTokenType     = prior[8]
        oxTokenValue    = prior[9]
        
    end if
    
end procedure


-----------------------------------------------------------------------------
global procedure oxUnloadBuffer()
    -- same as oxUnloadFile
    oxUnloadFile()
end procedure
                                                

-----------------------------------------------------------------------------
function oxExtractQuote( sequence string, sequence endOf, sequence escOf )
    -- returns quote from string, assumes to start from position 1
    -- if not found, issue an error
    integer i, char, theEsc, theEnd
    
    -- get chars
    theEsc = escOf[1]
    theEnd = endOf[1]

    -- start parsing past delimiter    
    i = 2
    while i <= length( string ) do 
        -- get char
        char = string[i]
        
        -- escape char?
        if char = theEsc then
            -- skip next char
            i+= 1            
            
        -- close quote?
        elsif char = theEnd then
            -- return string
            return string[1..i]
        end if
        
        -- move ahead
        i+= 1
        
    end while
    
    -- error
    oxParseError( "End of line reached with no closing %s", {endOf} )
    return string[1..length(string)]
    
end function


-----------------------------------------------------------------------------
function oxAtLiteral( sequence s )
    -- return true if non-keyword literal is at start of string
    sequence literal
    for i = 1 to length( oxKeyword ) do          
        -- get literal
        literal = oxKeyword[i]
        
        -- only check if doesn't start with a keyword char
        if length( literal ) < length( s )
        and not find( literal[1], oxInitKeyWordChars ) then
            -- equal?
            if equal( literal, s[1..length(literal)] ) then
                return i
            end if
        end if
    end for
    return 0
end function


-----------------------------------------------------------------------------
function oxParseDecimalNumber( sequence string )
    -- return decimal number from string
    -- handles exponents as well
    integer i, decimal, exponent, char
    sequence word
    
    word = {}                      
    decimal = False
    exponent = False

    i = 0
    while i < length( string ) do
    
        -- move ahead
        i += 1
        char = string[i]
    
        -- digit?
        if find( char, DecimalChars ) then
            word &= char
        
        -- decimal?
        elsif char = '.' then
            -- already encountered, or part of exponent?
            if decimal or exponent then
                -- end of number
                exit                
                
            -- not followed by a digit?
            elsif i = length(string)
            or    not find( string[i+1], DecimalChars ) then
                exit
            
            else
                -- add decimal point, set flag
                word &= char
                decimal = True
                
            end if
            
        -- exponent?
        elsif char = 'e'
        or    char = 'E' then
            -- since this routine is only called if the first char
            -- of the string is a decimal or decimal point, i don't
            -- have to check if the exponent is at the beginning of
            -- the string.
        
            -- already encountered?
            if exponent then
                -- end of number
                exit
            
            -- next char + or -
            elsif i+2 <= length( string )
            and   (string[i+1] = '+' or string[i+1] = '-')
            and   find( string[i+2], DecimalChars ) then
                -- 123e+3: valid exponent
                -- append and move ahead
                word &= string[i..i+1]
                i += 1
            
            -- next char decimal?        
            elsif i < length( string )
            and   find( string[i+1], DecimalChars ) then
                -- 123e3
                -- valid exponent
                word &= char
                
            else
                -- not valid exponent
                exit
            end if
            
            -- set flag
            exponent = True
            
        else
            exit
        end if
    end while

    return word
    
end function
        

-----------------------------------------------------------------------------
function oxStringToWords( sequence string )
    
    -- convert a sequence of chars into words
    
    integer at
    sequence words, word

    words = {}
    word = ""
    string &= " "

    while length( string ) do

        -- white space?
        if find( string[1], oxWhiteSpace ) then
            -- noise
            word = " "

        -- string?          
        elsif length( oxStringStart )
        and   match( oxStringStart, string ) = 1 then
            word = oxExtractQuote( string, oxStringEnd, oxStringEsc )

        -- char?
        elsif length( oxCharStart )
        and   match( oxCharStart, string ) = 1 then
            word = oxExtractQuote( string, oxCharEnd, oxCharEsc )

            -- check length
            if length( word ) = 4 then
                if word[2] != oxCharEsc[1] then
                    oxParseError( "Character constant %s too large", {word} )
                    exit
                end if
            elsif length( word ) != 3 then
                oxParseError( "Character constant %s too large", {word} )
                exit
            end if

        -- start of comment?
        elsif length( oxCommentStart )
        and   match( oxCommentStart, string ) = 1 then
            word = oxCommentStart

            -- in comment?
            if not oxInComment then
                -- set flag
                oxInComment = True
            end if
            
        -- end of comment?     
        elsif length( oxCommentEnd )
        and   match( oxCommentEnd, string ) = 1 then
            word = oxCommentEnd

        -- line comment?
        elsif not oxInComment
        and   length( oxSingleComment )
        and   match( oxSingleComment, string ) = 1 then
            -- all done
            exit
                      
        -- line comment style 2?
        elsif not oxInComment
        and   length( oxSingleComment2 )
        and   match( oxSingleComment2, string ) = 1 then
            -- all done
            exit
                      
        else
            -- literal?
            at = oxAtLiteral( string )
            if at then
                word = oxKeyword[at]

            -- not a literal?
            else
                           
                -- start of keyword?
                if find( string[1], oxInitKeyWordChars ) then
                    -- build keyword
                    word = {string[1]}
                    
                    -- scan through buffer
                    for i = 2 to length( string ) do
                    
                        -- non-keyword char or embedded token?
                        if not find( string[i], oxKeyWordChars )
                        or oxAtLiteral( string[i..length(string)] ) then
                            -- end of match
                            exit
                        else
                            -- accumulate
                            word &= string[i]
                        end if
                    end for
            
                -- decimal digit?
                elsif find( string[1], DecimalChars ) then
                    word = oxParseDecimalNumber( string )

                -- decimal point?
                elsif 2 < length( string )
                and string[1] = '.'
                and find( string[2], DecimalChars ) then
                    -- parse number
                    word = oxParseDecimalNumber( string )            


                -- hexidecimal number?                
                elsif length( oxHexStart )
                and   match( oxHexStart, string ) = 1 then
                    -- parse hex number
                    word = oxHexStart
                    for i = length(oxHexStart)+1 to length( string ) do
                        if find( string[i], HexChars ) then
                            word &= string[i]
                        else
                            exit
                        end if
                    end for     
                    
                else
                    -- treat char as word
                    word = {string[1]}
                    
                end if
            end if
        end if

        -- remove word from string
        string = string[length(word)+1..length(string)]

        -- accumulate?
        if not oxInComment
        and not equal( word, " " ) then
            words &= {word}
        end if


        -- turn off comment flag?
        if equal( word, oxCommentEnd ) then
            oxInComment = False
        end if
                    
            
    end while

-- for debugging the lexer
-- for i = 1 to length( words ) do
--     printf( 1, "%d %s\n", {i,words[i]} )
-- end for
    
    return words

end function

-----------------------------------------------------------------------------
procedure oxNextLineGet()

    -- convert the next line of text in oxText into
    -- a string of tokens

    -- end of document?
    if oxLineNum >= length( oxStrings ) then
        -- end of file is current line
        oxString = "<EOF>"
        
        -- set as single token returned
        oxTokens = { oxString }
    
    else
        -- increment line number
        oxLineNum += 1
        
        -- set current line
        oxString = oxStrings[oxLineNum]
        
        -- parse current line
        oxTokens = oxStringToWords( oxString )

    end if

    -- debugging
    -- puts( 1, oxString )
    
end procedure


-----------------------------------------------------------------------------
global function oxLineGet()
    -- returns contents of current line,
    -- moves to next line
    sequence line
    line = oxString
    oxNextLineGet()
    oxTokenIndex = 1   
    return line
end function


-----------------------------------------------------------------------------
procedure oxMoveAhead()
    -- get the next token from the buffer
    oxTokenIndex += 1
    oxToken = ""
end procedure


-----------------------------------------------------------------------------
function oxCurrentToken()

    -- need to parse a new line of tokens?
    if length( oxTokens ) < oxTokenIndex
    or oxTokenIndex = 0 then
        -- clear token list and index to current token
        oxTokens = {}      
        oxTokenIndex = 1   
        
        -- read until tokens are accumulated
        while length(oxTokens) = 0 do
            -- get the next line from the strings, and tokenize
            oxNextLineGet()
        end while
    end if    

    -- new line start?
    if oxLineId = 0 then
        oxLineId = oxLineNum
    end if
    
    -- no current token?
    if length( oxToken ) = 0 then
        oxEvaluateWord( oxTokens[oxTokenIndex] )
    end if

    return oxTokenType
    
end function


-----------------------------------------------------------------------------
global function oxTokenGet()
    -- return current token,
    -- move ahead
    sequence token, word
    word = oxCurrentToken()
    token = oxToken
    oxMoveAhead()
    return { token, word }
end function
                         
-----------------------------------------------------------------------------
global function oxTokenPeek()
    -- return current token, 
    -- don't move ahead
    sequence token, word
    word = oxCurrentToken()
    token = oxToken
    return { token, word }
end function
                         

-----------------------------------------------------------------------------
global function oxTokenIs( sequence word )
    if oxIgnoreCase then
        return equal( upper(oxCurrentToken()), word )
    else
        return equal( oxCurrentToken(), word )
    end if
end function

-----------------------------------------------------------------------------
global function oxEatToken( sequence word )
    if oxTokenIs( word ) then
        oxMoveAhead()
        return True
    else
        return False        
    end if
end function

-----------------------------------------------------------------------------
global procedure oxExpected( sequence list )

    -- display an error message, showing the expected tokens.
    sequence msg, item

    -- skip if error already generated
    if oxError then
        return
    end if

    -- avoid overly long error messages
    if length( list ) > 9 then
        list = {}
    end if

    if length( list ) then
        -- sort the list
        list = sort( list )
    
        -- build a message with a list of possible options
        msg = "expected "
        
        -- iterate through list
        for i = 1 to length( list ) do
            
            -- get the item
            item = list[i]
            if equal( item, "NUM" ) then
                item = "a number"
                
            elsif equal( item, "STR" ) then
                item = "a string"

            elsif equal( item, "CHAR" ) then
                item = "a character"

            elsif equal( item, "ID" ) then
                item = "an identifier"

            else
                -- quote it
                item = "'" & item & "'"
                
            end if
            
            -- end of list?
            if i = length( list ) then
                -- only item?
                if i = 1 then
                    msg &= item
                else
                    -- end of list
                    msg &= " or " & item
                end if       
                
            -- next to last item?
            elsif i = length( list ) -1 then
                msg &= item
                
            -- not at end, so add comma
            else
                msg &= item & ", "
            end if
        end for
        
        msg &= ", not "

    else
        -- can't offer help
        msg = "unexpected input: "
        
    end if  

    -- need to quote token?
    if equal( oxToken, "<EOF>" ) then
        -- no quotes
        msg &= "%s"
    else
        msg &= "'%s'"
    end if

    -- write error
    oxParseError( msg, {oxToken} )
    
end procedure


-----------------------------------------------------------------------------
global function oxExpect( sequence word )
    -- the word must match the token.
    -- if it does, returns the token's value.
    -- if not, generates an error
    if oxEatToken( word ) then
        return oxTokenValue
    else
        oxExpected( {word} )
        return False
    end if
end function


-----------------------------------------------------------------------------
global procedure oxScanUntil( sequence word )
    -- skip all tokens until token match
    while not oxTokenIs( word ) do
        if oxTokenIs( "<EOF>") then
            oxParseError( "Expected \"%s\", not \"%s\"", {word, oxToken} )
            return
        else
            oxMoveAhead()
        end if
    end while
end procedure



-----------------------------------------------------------------------------
global function oxParse_( sequence rule )
    sequence rules, code, expected, tokens
    integer op, count

    -- get the rules
    rules = oxParseRules
    
    -- clear count
    count = 0

    -- clear the parsed tokens list
    tokens = {}              

    while 1 do                     
        -- clear the expected tokens list
        expected = {}

        -- run the rules        
        for i = 1 to length( rules ) do
            code = rules[i]
            op = code[1]

            if (op = OxRule and equal( rule, code[2] ) )
            or (op = OxToken and oxTokenIs( code[2] )  )
            or op = OxTrue then   

-- trace the parse
-- if op = OxRule then
--    printf( 1, "lex: %s ", {rule} )
-- end if


                -- rule match, take branch
                if op = OxTrue then
                    -- no ops
                    rules = code[2..length(code)]
                else
                    -- one op
                    rules = code[3..length(code)]
                end if
                
                -- leave loop
                exit

            elsif op = OxToken then
                -- failed first test, add to exected list
                expected &= {code[2]}

            elsif op = OxRule
            or    op = OxTrue then
                -- do nothing, failed above

            elsif op = OxExpect then
                -- accumulate token
                tokens &= {oxExpect( code[2] )}

            elsif op = OxParse then
                tokens &= {oxParse_( code[2] )}
            
            elsif op = OxPre then                      
                -- call the routine and continue
                call_proc( code[2], {tokens} )

            elsif op = OxCall then                      
                -- call the routine and exit
                return call_func( code[2], {tokens} )

            elsif op = OxTerminal then
                -- true, but no rule to execute
                return {}
                         
            elsif op = OxFailed then     
                -- error message
                oxExpected( expected )

            else
                -- failed     
                printf( 1, "Parser Internal Error: unknown opcode %d\n", {op} )
                oxError = 1
                return {}
                
            end if

            -- error?
            if oxError then
                return {}
            end if

        end for

        count += 1
        if count > 500 then
            -- failed     
            printf( 1, "Parser Internal Error: infinite loop in %s\n", {rule} )
            oxError = 1
            return {}
        end if
        
    end while
    
end function

-----------------------------------------------------------------------------
global function oxParse( sequence rule )
    -- clear error flag and call the parser
    oxError = 0
    return oxParse_( rule )
end function


-----------------------------------------------------------------------------
global procedure oxInit()              

    -- called before running the parser

    -- initialize state
    oxFileName      = ""
    oxInComment     = False
    oxLineNum       = 0
    oxLineId        = 0
    oxPrior         = {}
    oxString        = ""
    oxStrings       = {}
    oxToken         = ""
    oxTokenIndex    = 0
    oxTokenType     = ""
    oxTokenValue    = 0
    oxTokens        = {}
    oxError         = False

    -- whatever whitespace chars are not literals are added
    -- to the list of whitespace characters
    oxWhiteSpace = " "
    for i = 1 to length( WhiteSpace ) do     
        -- not literal?
        if not find( {WhiteSpace[i]}, oxKeyword ) then
            oxWhiteSpace &= WhiteSpace[i]
        end if
    end for
    
end procedure

